#include <algorithm>
#include <cassert>
#include <iostream>
using std::cout; using std::endl;

#include "Fold.hpp"
#include "Thread.hpp"

using std::vector;

#define NTHREADS (4)

Learner::~Learner()
{
}

Fold::Fold()
{
}

Fold::Fold(const vector<QAPSum> &train, const vector<QAPSum> &test) :
  train(train), test(test)
{
}

Fold::~Fold()
{
}

/**
 * The number of incorrect elements in the permutation
 */
static int permWrong(const Permutation &P, const Permutation &truth)
{
  assert(P.n == truth.n);

  int nwrong = 0;
  for(int i = 0; i < P.n; i++) {
    if(P[i] != truth[i]) {
      nwrong++;
    }
  }
  return nwrong;
}

vector<TestResult> Fold::run(Learner &learner) const
{
  // Start fresh
  learner.reset();

  // Start with training
  std::vector<Permutation> truth;
  truth.reserve(train.size());
  for(int i = 0; i < train.size(); i++) {
    truth.push_back(Permutation(train[i].n));
  }

  

  learner.train(train, truth);

  // Test
  std::vector<TestResult> results(test.size());
  for(int i = 0; i < test.size(); i++) {
    results[i].truth = Permutation(test[i].n);
    results[i].predict = learner.predict(test[i], &results[i].nvisited);
  }

  return results;
}

vector<Fold> Fold::from_shuffle(const vector<QAPSum> &all, int n, const vector<int> &shuffle)
{
  assert(all.size() == shuffle.size());
  int N = all.size();

  vector<Fold> folds(n);
  
  for(int i = 0; i < n; i++) {
    int test_begin = N/n * i;
    int test_end = N/n * (i+1);
    if(i == n-1) {
      test_end = N;
    }

    // Copy into fold i
    // TODO: use shuffled
    Fold &F = folds[i];

    F.train.reserve(N - (test_end - test_begin));
    F.train_ind.reserve(N - (test_end - test_begin));
    for(int j = 0; j < test_begin; j++) {
      assert(shuffle[j] >= 0 && shuffle[j] < N);
      F.train.push_back(all[shuffle[j]]);
      F.train_ind.push_back(shuffle[j]);
    }
    for(int j = test_end; j < N; j++) {
      assert(shuffle[j] >= 0 && shuffle[j] < N);
      F.train.push_back(all[shuffle[j]]);
      F.train_ind.push_back(shuffle[j]);
    }

    F.test.reserve(test_end - test_begin);
    F.test_ind.reserve(test_end - test_begin);
    for(int j = test_begin; j < test_end; j++) {
      assert(shuffle[j] >= 0 && shuffle[j] < N);
      F.test.push_back(all[shuffle[j]]);
      F.test_ind.push_back(shuffle[j]);
    }
  }

  return folds;
}

vector<Fold> Fold::nfold(const vector<QAPSum> &all, int n, std::ostream *fold_file)
{
  int N = all.size();

  // Shuffle dataset
  vector<int> shuffle(N);
  for(int i = 0; i < N; i++) {
    shuffle[i] = i;
  }
  random_shuffle(shuffle.begin(), shuffle.end());

  // Write to file
  if(fold_file != NULL) {
    for(int i = 0; i < N; i++) {
      (*fold_file)<<shuffle[i]<<' ';
    }
  }

  return from_shuffle(all, n, shuffle);
}

vector<Fold> Fold::from_file(const vector<QAPSum> &all, int n, std::istream &fold_file)
{
  // Read in shuffle
  int N = all.size();
  vector<int> shuffle(N);
  for(int i = 0; i < N; i++) {
    fold_file>>shuffle[i];
  }

  return from_shuffle(all, n, shuffle);
}

class FoldThread: public Thread
{
public:
  const vector<Fold> &folds;
  int begin, end;
  vector<vector<TestResult> > &results;
  Learner *learner;

  FoldThread(const vector<Fold> &init_folds, int init_begin, int init_end, const Learner &base_learner, vector<vector<TestResult> > &init_results) :
    folds(init_folds), begin(init_begin), end(init_end), results(init_results)
  {
    learner = base_learner.clone();
  }

  ~FoldThread()
  {
    delete learner;
  }

  virtual void run()
  {
    for(int i = begin; i < end; i++) {
      results[i] = folds[i].run(*learner);

      // Print fold summary
      cout<<"Fold "<<(i + 1)<<endl;
      cout<<"  accuracy: "<<TestResult::accuracy(results[i].begin(), results[i].end())<<endl;
      cout<<"  nvisited: "<<TestResult::averageVisited(results[i].begin(), results[i].end())<<endl;
    }
  };

private:
  FoldThread(const FoldThread &);
  const FoldThread &operator = (const FoldThread &);
};

vector<vector<TestResult> > Fold::run(const vector<Fold> &F, Learner &learner, int nthreads)
{
  vector<vector<TestResult> > results(F.size());
  FoldThread *threads[nthreads];

  // Create threads
  for(int i = 0; i < nthreads; i++) {
    int begin = (F.size() / nthreads) * i;
    int end = (F.size() / nthreads) * (i + 1);
    if(i == nthreads - 1) {
      end = F.size();
    }

    threads[i] = new FoldThread(F, begin, end, learner, results);
  }

  for(int i = 0; i < nthreads; i++) {
    threads[i]->start();
  }

  // Wait for completion
  for(int i = 0; i < nthreads; i++) {
    threads[i]->join();
    delete threads[i];
  }

  // Print overall summary
  vector<TestResult> all = mergeResults(results);
  cout<<"Overall"<<endl;
  cout<<"  accuracy: "<<TestResult::accuracy(all.begin(), all.end())<<endl;
  cout<<"  nvisited: "<<TestResult::averageVisited(all.begin(), all.end())<<endl;

  return results;
}

std::vector<TestResult> Fold::mergeResults(const std::vector<std::vector<TestResult> > &fold_results)
{
  // Allocated merged vector
  int n = 0;
  for(int i = 0; i < fold_results.size(); i++) {
    n += fold_results[i].size();
  }

  std::vector<TestResult> merged;
  merged.reserve(n);

  // Assemble merged vector
  for(int i = 0; i < fold_results.size(); i++) {
    for(int j = 0; j < fold_results[i].size(); j++) {
      merged.push_back(fold_results[i][j]);
    }
  }

  return merged;
}
