#include <cmath>
#include <iostream>
using std::cout; using std::endl;

#include "QAP/QAP.hpp"
#include "utility/Format.hpp"

int branchCounter;

static const int n = 12;

/**
 * Load the edge graph for house @c i
 */
static Graph loadGraph(int i)
{
  std::string path = stringFormat("graphmatchBMRM/Data/house/house%03d.adj", i);
  //cout<<"Loading "<<path<<"..."<<endl;
  return Graph(path);
}

/**
 * Load the shape context features for house @c i
 */
static Fmatrix loadSCF(int i)
{
  std::string path = stringFormat("graphmatchBMRM/Data/house/house%03d.scf", i);
  //cout<<"Loading "<<path<<"..."<<endl;
  return Fmatrix(path);
}

static double square(double x)
{
  return x*x;
}

/**
 * Load the shape context features for house @c i
 */
static Graph makePhi(Fmatrix A, Fmatrix B)
{
  assert(A.Ncols == B.Ncols);
  assert(A.Nrows == B.Nrows);

  Graph Phi(A.Nrows);
  for(int i = 0; i < A.Nrows; i++) {
    for(int j = 0; j < B.Nrows; j++) {
      // Phi(i, j) = distance(A(i, :), B(j, :))

      double dist = 0;
      for(int k = 0; k < A.Ncols; k++) {
	dist += square(A(i, k) - B(j, k));
      }
      dist = sqrt(dist);

      Phi(i, j) = dist;
    }
  }

  return Phi;
}



/**
 * Graph the first n nodes of a graph.
 */
static Graph truncateGraph(int n, const Graph &G)
{
  Graph T(n);

  for(int i = 0; i < n; i++) {
    for(int j = 0; j < n; j++) {
      T(i, j) = G(i, j);
    }
  }

  return T;
}

/**
 * Load the QAP between house @i and house @j
 */
static QAP loadQAP(int i, int j)
{
  Graph G1 = loadGraph(i);
  Fmatrix scf1 = loadSCF(i);

  Graph G2 = loadGraph(j);
  Fmatrix scf2 = loadSCF(j);

  //NIPS submission cost function
  // Graph Phi = makePhi(scf1, scf2);
  // Phi = Phi.pseudoNegate();
  
  // Phi = truncateGraph(n, Phi);


  //ICML cost function
  Graph k1 = makePhi(scf1, scf1);
  Graph k2 = makePhi(scf2, scf2);

  G1 = truncateGraph(n, G1);
  G2 = truncateGraph(n, G2);
  k1 = truncateGraph(n, k1);
  k2 = truncateGraph(n, k2);

 


// cout<<"G1 = "<<endl<<G1<<endl;
 //  cout<<"G2 = "<<endl<<G2<<endl;
 //  cout<<"Phi = "<<endl<<Phi<<endl;
 //  getchar();

  //return QAP(G1, G2);
  // return QAP(G1, G2, Phi);
  return QAP(G1, G2, k1, k2);

}

/**
 * The number of elements of the permutation that do not map to themselves.
 */
static int permWrong(const Permutation &P)
{
  int nwrong = 0;
  for(int i = 0; i < P.n; i++) {
    if(P[i] != i+1) {
      nwrong++;
    }
  }
  return nwrong;
}

int main(int argc, char **argv)
{
  QAP qap = loadQAP(11, 21);
  WTree w(qap);
  Permutation truth(n);

  int nwrong = 0;
  for(int i = 41; i <= 50; i++) {
    QAP testqap = loadQAP(i, i+10);

    Permutation P = testqap.solve(&w);
    cout<<P<<endl;

    nwrong += permWrong(P);
  }
  cout<<"Before training: "<<nwrong<<endl;

  for(int i = 1; i <= 40; i++) {
    QAP qap = loadQAP(i, i+10);
    qap.train(truth, &w);

    cout<<"nexpanded = "<<w.nexpanded()<<endl;
  }


  std::ofstream wfile("w.param", std::ios::out | std::ios::trunc);
  w.write(wfile);

  nwrong = 0;
  for(int i = 41; i <= 50; i++) {
    QAP testqap = loadQAP(i, i+10);

    Permutation P = testqap.solve(&w);
    cout<<P<<endl;

    nwrong += permWrong(P);
  }
  cout<<"After training: "<<nwrong<<endl;

  return 0;
}
