#include <cmath>

#include "ImageSequence.hpp"

#include "utility/Format.hpp"

/**
 * Load the edge graph for @c i
 */
static Graph loadGraph(const std::string &format, int i)
{
  std::string path = stringFormat(format.c_str(), i);
  return Graph(path);
}

 
/**
 * Load the shape context features for @c i
 */
static Fmatrix loadSCF(const std::string &format, int i)
{
  std::string path = stringFormat(format.c_str(), i);
  return Fmatrix(path);
}
 
/**
 * Load the node co-ordinates for @c i
 */
static Fmatrix loadPoints(const std::string &format, int i)
{
  std::string path = stringFormat(format.c_str(), i);
  return Fmatrix(path);
}

static double square(double x)
{
  return x*x;
}

/**
 * Create a graph of pairwise distances from a pair of SCF matrices
 */
static Graph makePhi(Fmatrix A, Fmatrix B)
{
  assert(A.Ncols == B.Ncols);
  assert(A.Nrows == B.Nrows);

	Graph Phi(A.Nrows);
	for(int i = 0; i < A.Nrows; i++) {
	  for(int j = 0; j < B.Nrows; j++) {
	    
	    double dist = 0;
	    for(int k = 0; k < A.Ncols; k++) {
	      dist += square(A(i, k) - B(j, k));
	    }
     
	    dist = sqrt(dist);
	    
	    Phi(i, j) = dist;
	  }
	}

 // cout<<"Phi "<<endl<<Phi<<endl;
 //  getchar();
  return Phi;
}




/**
 * Create a graph of pairwise distances from a pair of noisy SCF matrices
 */
static Graph makePhi_Noise(Fmatrix A, Fmatrix B)
{
  assert(A.Ncols == B.Ncols);
  assert(A.Nrows == B.Nrows);

	Graph Phi(A.Nrows);
	for(int i = 0; i < A.Nrows; i++) {
	  for(int j = 0; j < B.Nrows; j++) {
	    
	    double dist = 0;
	    for(int k = 0; k < A.Ncols; k++) {
	      dist += square(A(i, k) - B(j, k));
	    }
     
	    dist = sqrt(dist);
	    
	    Phi(i, j) = dist;
	  }
	}

 // cout<<"Phi "<<endl<<Phi<<endl;
 //  getchar();
  return Phi;
}




/**
 * Create a graph of pairwise distances from a pair of SCF matrices
 */
static Graph makePhiExp(Fmatrix A, Fmatrix B, double s)
{
  assert(A.Ncols == B.Ncols);
  assert(A.Nrows == B.Nrows);

  Graph Phi(A.Nrows);
   for(int i = 0; i < A.Nrows; i++) {
    for(int j = 0; j < B.Nrows; j++) {
      // Phi(i, j) = distance(A(i, :), B(j, :))

      double dist = 0;
      for(int k = 0; k < A.Ncols; k++) {
	dist += square(A(i, k) - B(j, k));
      }
      dist = exp(-dist/2*s*s);

      Phi(i, j) = dist;
    }
  }

  return Phi;
}


/**
 * Build the set of QAPs matching image i and j for a dataset as specified.
 */
static QAPSum buildQAPs(
  int i, int j,
  const std::string &delaunay_format, const std::string &scf_format, const std::string &points_format)
{
  std::vector<QAP> qq;
  
  // QAP based on matching Delaunay triangulations
  Graph delaunay1 = loadGraph(delaunay_format, i);
  Graph delaunay2 = loadGraph(delaunay_format, j);

  qq.push_back(QAP(delaunay1, delaunay2, "delaunay"));

  // Distance-based QAPs
  Fmatrix points1 = loadPoints(points_format, i);
  Fmatrix points2 = loadPoints(points_format, j);

  double scale = 1;
  for(int k = 0; k < 5; k++) {
    Graph dists1 = makePhiExp(points1, points1, scale);
    Graph dists2 = makePhiExp(points2, points2, scale);

    qq.push_back(QAP(dists1, dists2, stringFormat("distance (sigma = %.1lg)", scale)));

    scale /= 10;
  }

  // Junk QAPs
  int n = qq[0].n;
  for(int p = 0; p < 5; p++) {
    Graph::RandomNedges k1(n, 10);
    Graph::RandomNedges k2(n, 10);

    qq.push_back(QAP(k1, k2, "noise"));
  }

  // Shape context features (SCF)
  Fmatrix scf1 = loadSCF(scf_format, i);
  Fmatrix scf2 = loadSCF(scf_format, j);
  
  vector<int> row(scf1.Nrows);
  for(int k = 0; k < row.size(); k++){
    row[k]=k;
  }
  vector<int> col(20);
  assert(scf1.Ncols == scf2.Ncols);//number of features are same in the two graphs
  for(int k = 0; k < scf1.Ncols;){
    for(int c = 0; c < col.size(); c++){
      col[c]=c+k;
    }
   
      Graph G1 = makePhi(scf1.matrixMinor(row,col), scf1.matrixMinor(row,col));
      Graph G2 = makePhi(scf2.matrixMinor(row,col), scf2.matrixMinor(row,col));
      //qq.push_back(QAP(G1, G2, stringFormat("scf %d", k)));
     
    k = k+col.size();
  }

  // Check sizes + return result
  for(int k = 1; k < qq.size(); k++) {
    assert(qq[k].n == qq[0].n);
  }
  return QAPSum(qq);
}


std::vector<QAPSum> imageSequence(
  int n, int baseline,
  const std::string &delaunay_format, const std::string &scf_format, const std::string &points_format)
{
  std::vector<QAPSum> seq(n-baseline);

  for(int i = 1; i <= n - baseline; i++) {
    seq[i-1] =  buildQAPs(i, i + baseline, delaunay_format, scf_format, points_format);
  }

  return seq;
}
