#include <cstdlib>
#include <cmath>
extern "C" {
#include <memory.h>
}

#include "QAPSum.hpp"
#include "QAPbranch.hpp"

#ifndef INFINITY
#define INFINITY (1.0/0.0)
#endif

//////////////////////////// Weights //////////////////////////

QAPSum::Weights::Weights(int _m) :
  m(_m), w(new double[_m])
{
  for(int i = 0; i < m; i++) {
    w[i] = 1.0;
  }
  //cout<<"w "<<w.length()<<endl;
}

QAPSum::Weights::Weights(int _m, const double *_w) :
  m(_m), w(new double[_m])
{
  for(int i = 0; i < m; i++) {
    w[i] = _w[i];
  }
}


QAPSum::Weights::Weights(const Weights &copy) :
  m(copy.m), w(new double[copy.m])
{
  for(int i = 0; i < m; i++) {
    w[i] = copy.w[i];
  }
}

QAPSum::Weights::~Weights()
{
  delete [] w;
}

const QAPSum::Weights &QAPSum::Weights::operator = (const Weights &assign)
{
  if(m != assign.m) {
    delete [] w;

    m = assign.m;
    w = new double[m];
  }

  for(int i = 0; i < m; i++) {
    w[i] = assign.w[i];
  }

  return *this;
}

int QAPSum::Weights::length() const
{
  return m;
}

double *QAPSum::Weights::weights()
{
  return w;
}

const double *QAPSum::Weights::weights() const
{
  return w;
}

double QAPSum::Weights::norm() const
{
  double S = 0;
  for(int i = 0; i < m; i++) {
    S += w[i] * w[i];
  }
  return sqrt(S);
}

const QAPSum::Weights &QAPSum::Weights::operator += (const Weights &addend)
{
  assert(m == addend.m);
  for(int i = 0; i < m; i++) {
    w[i] += addend.w[i];
  }
  return *this;
}

const QAPSum::Weights &QAPSum::Weights::operator -= (const Weights &subtrahend)
{
  assert(m == subtrahend.m);
  for(int i = 0; i < m; i++) {
    w[i] -= subtrahend.w[i];
  }
  return *this;
}

const QAPSum::Weights &QAPSum::Weights::operator *= (double s)
{
  for(int i = 0; i < m; i++) {
    w[i] *= s;
  }
  return *this;
}

const QAPSum::Weights &QAPSum::Weights::operator /= (double s)
{
  for(int i = 0; i < m; i++) {
    w[i] /= s;
  }
  return *this;
}


std::ostream &operator << (std::ostream &stream, const QAPSum::Weights &w)
{
  int m = w.length();
  assert(m >= 0);

  stream<<'[';
  if(m != 0) {
    stream<<w[0];
  }
  for(int i = 1; i < m; i++) {
    stream<<", "<<w[i];
  }
  stream<<']';

  return stream;
}

////////////////////////////////// QAPSum /////////////////////////

QAPSum::QAPSum() :
  m(0)
{
}

QAPSum::QAPSum(const std::vector<QAP> &_Q) :
  m(_Q.size()), n(_Q[0].n), Q(_Q)
{
}

QAPSum::~QAPSum()
{
}

Permutation QAPSum::solve(const Weights &w, int *nvisited, bool greedy, int max_visit) const
{
  assert(w.length() == m);

  // Run branch-and-bound
  QAPBnB bnb(*this, max_visit);
  bnb.greedy = greedy;

  QAPbranch root(bnb, *this);

  double obj = root.BranchAndBound(-INFINITY, w);


  // Save the count of visited nodes
  if(nvisited != NULL) {
    *nvisited = bnb.nvisited;
  }

  // Generate true permutation from bestseq
  return Permutation(CClist(bnb.bestseq)).inverse();
}


// static void objectiveIFT(QAPbranch *branch, FIELD *f)
// {
//   const Fourier &F = *branch->Fmod;
//   const FourierT &T = F.T;

//   size_t n = T.n;
//   size_t order = (size_t)T.order;
//   size_t suborder = order/n;
//   assert(order % n == 0);

//   // Trivial function over S_1 is identical to its fourier transform
//   if(order == 1){
//     f[0] = F[0]->at(0,0);
//     return;
//   }
  
//   // Iterate over individual examples
//   for(int j = 1; j <= n; j++){
//     QAPbranch *sub = branch->makeBranch(j);
//     assert(sub->Fmod->T.order == suborder);

//     objectiveIFT(sub, f + (n-j)*suborder);

//     delete sub;
//   }
// }

// SnSF::Function QAP::objectiveFunction(WTree *w)
// {
//   ///*
//   QAPBnB bnb(*this, w);
//   QAPbranch root(bnb, G1, G2, k1, k2, w);

//   SnSF::Function func(unary ? (*root.Fmod) : (*root.F));
//   SnSF::Function result(w->getT());
//   memcpy(result.f, func.f, result.T.order * sizeof(FIELD)); 
//   return result;
//   //*/

//   /*
//   // Generate BnB tree
//   QAPBnB bnb(*this, w);
//   QAPbranch root(bnb, G1, G2, phi, w);

//   // Perform IFT
//   FIELD *f = new FIELD[(size_t)(root.Fmod->T.order)];
//   objectiveIFT(&root, f);

//   assert(w->getT().order == root.Fmod->T.order);

//   // Return function
//   SnSF::Function func(w->getT(), f);
//   func.protect = 0;
//   return func;
//   //*/
// }

void QAPSum::permuteProblem(const Permutation &P)
{
  for(int i = 0; i < m; i++) {
    Q[i].permuteProblem(P);
  }
}
