#include "QAPbranch.hpp"
#include "QAP.hpp"
#include "maxFourier.hpp"

//#include "Matlab.hpp"
#include "BlockMatrix.hpp"
#include "Timer.hpp"

#include <cmath>
#include <algorithm>
#include <sstream>
#include <cassert>

/////////////////////////// QAPBnB /////////////////////////

QAPBnB::QAPBnB(const QAPSum &_QAPs, int init_max_visit) :
  m(_QAPs.m), n(_QAPs.n), QAPs(_QAPs),
  learning(false),
  greedy(false),
  maxsofar(0),
  nvisited(0),
  max_visit(init_max_visit)
{
  // Graph function on adjacency matices
  Tgraph = new Transform::Left(n, n-2);

  // Tcorr: more general transform not restricted to homogeneous space
  Tcorr = new Transform::Left(OuterProduct<Transform>(*Tgraph, *Tgraph));
}

QAPBnB::QAPBnB(const QAPSum &_QAPs, const Permutation &_truth) :
  m(_QAPs.m), n(_QAPs.n), QAPs(_QAPs), 
  learning(true), truth(_truth), true_cclist(_truth.inverse().cclist()),
  greedy(false),
  maxsofar(0),
  nvisited(0),
  max_visit(-1)
{
  // Graph function on adjacency matices
  Tgraph = new Transform::Left(n, n-2);

  // Tcorr: more general transform not restricted to homogeneous space
  Tcorr = new Transform::Left(OuterProduct<Transform>(*Tgraph, *Tgraph));

  // Get the offset needed to negate the objective function
  // TODO
  /*
  owner.negationoffset=0;
  for(int i=0; i<n; i++) {
    for(int j=0; j<n; j++) {
      if(i!=j) {
	owner.negationoffset += max*G1(i,j);
      }
    }
  }
  */
}

QAPBnB::~QAPBnB()
{
  delete Tgraph;
  delete Tcorr;
}

/////////////////////////// QAPbranch //////////////////////////

QAPbranch::QAPbranch(QAPBnB &_owner, int _m, int _n) :
  owner(_owner), m(_m), n(_n),
  F(_m), bounds(NULL)
{
  //cout<<"Creating branch at depth "<<depth()<<endl;
}

QAPbranch::QAPbranch(QAPBnB &_owner, const QAPSum &QAPs) :
  owner(_owner), m(owner.m), n(owner.n),
  F(owner.m), bounds(NULL)
{
 
  // Calculate products
  for(int i = 0; i < m; i++) {
    Fourier *F1 = QAPs.Q[i].G1.Fourier(*owner.Tgraph);
    Fourier *F2 = QAPs.Q[i].G2.Fourier(*owner.Tgraph);
  
    // Kondor's Fourier QAP objective F = F1 * F2' 
    Product<Fourier, Transpose<Fourier> > prod(*F1, ~(*F2));
    F[i] = new Fourier(owner.Tcorr->top(), prod);
    
    delete F1;
    delete F2;
  }
}

QAPbranch::~QAPbranch()
{
  delete [] bounds;
  for(int i = 0; i < F.size(); i++) {
    delete F[i];
  }
}

bool QAPbranch::isLeaf() const
{
  assert(n >= 1);
  return n == 1;
}

int QAPbranch::depth() const
{
  return owner.n - n;
}

int QAPbranch::trueChild() const
{
  int depth = owner.n - n;
  assert(owner.true_cclist.size() >= depth + 1);
  return owner.true_cclist[owner.true_cclist.size() - 1 - depth];
}

QAPbranch* QAPbranch::makeBranch(int jj) const
{

  QAPbranch *sub = new QAPbranch(owner, m, n-1);
  const FourierT &subT = *(F[0]->T.descendant);

  for(int i = 0; i < m; i++) {
    // Take F inherited from parent
    Fourier subtildeF(*F[i]);
  
    subtildeF.rotateLeft(jj, 1);
    sub->F[i] = new Fourier(subT, subtildeF);
}
  
  return sub;
}

double QAPbranch::bound(const QAPSum::Weights &w)
{
  // Lazily construct bounds vector
  if(bounds == NULL) {
    bounds = new double[m];

    if(isLeaf()) {
      for(int i = 0; i < m; i++) {
	bounds[i] = (*F[i])[0]->at(0,0);
      }
    }
    else {
      for(int i = 0; i < m; i++) {
	// We replace Kondor's bound with this one which is proportional within
	// a given level of the tree.
	double bndi = 0;
	int Fsize = F[i]->size();
	for(int j = 0; j < Fsize; j++) {
	  bndi += F[i]->at(j)->Nrows * F[i]->at(j)->tracenorm();
	}
	bounds[i] = bndi;
      }
    }
  }

  double B = 0;
  for(int i = 0; i < m; i++) {
    B += w[i] * bounds[i];
  }
  return B;
}

std::vector<QAPbranch *> QAPbranch::makeChildren() const
{
   assert(!isLeaf());

  vector<QAPbranch *> children(n);
  for(int i = 0; i < n; i++) {
    children[i] = makeBranch(i+1);
  }

  return children;
}

static bool pairCompare(pair<int,double> a, pair<int,double> b)
{
  return (a.second > b.second);
}

double QAPbranch::BranchAndBound(double maxsofar, const QAPSum::Weights &w)
{
  // Increment node count
  owner.nvisited++;
  //if(owner.nvisited % 1000 == 0) {
  //  cout<<"nvisited = "<<owner.nvisited<<endl;
  //}
  if(owner.max_visit > 0 && owner.nvisited > owner.max_visit) {
    return maxsofar;
  }

  if(isLeaf()) {
    double val = bound(w);

    if(val > maxsofar){
      maxsofar = val; 
      owner.bestseq.clear();
      owner.bestseq.push_back(1);
    }

    return maxsofar;
  }

  // Create branches
  vector<QAPbranch *> branch = makeChildren();

  // Calculate bounds
  vector<pair<int, double> > norms;
  for(int i = 0; i < n; i++){
    norms.push_back(pair<int, double>(i, branch[i]->bound(w)));
  }
  sort(norms.begin(), norms.end(), pairCompare);

  bool winner = 0;
  
  if(owner.greedy) {
    // No backtracking, accepts first leaf visited
    maxsofar = branch[norms[0].first]->BranchAndBound(maxsofar, w);
    owner.bestseq.push_back(norms[0].first + 1);
  }
  else {
    // Normal branch-and-bound
    for(int i = 0; i < n && norms[i].second > maxsofar; i++) {
      double proposed = branch[norms[i].first]->BranchAndBound(maxsofar, w);
    
      if(proposed > maxsofar) {
        maxsofar = proposed;
        owner.bestseq.push_back(norms[i].first+1);
      }
    }
  }

  // Cleanup
  for(int i=0; i<branch.size(); i++) {
    delete branch[i];
  }
  	
  return maxsofar;
}
