#ifndef __QAPbranch_HPP__
#define __QAPbranch_HPP__

#include "SnSFclasses.hpp"
#include "QAPSum.hpp"
#include "AssignmentMatrix.hpp"
#include "Graph.hpp"

using namespace SnSF;

/**
 * A class setting up data for a QAP branch-and-bound learning
 */
class QAPBnB
{
public:
  /**
   * Initialize a BnB tree for simply solving without training.
   */
  QAPBnB(const QAPSum &QAPs, int max_visit = -1);

  /**
   * Initialize a BnB tree for learning parameters.
   */
  QAPBnB(const QAPSum &QAPs, const Permutation &truth);

  ~QAPBnB();

  int n, m;
  QAPSum QAPs;

  /**
   * Whether to perform parameter adjustments during BnB
   */
  bool learning;

  /**
   * Whether to perform backtracking
   */
  bool greedy;

  /**
   * Count the number of nodes used in a BnB run
   */
  int nvisited;
  
  /**
   * Set the maximum number of nodes to visit.
   *
   * @note If this is set to a nonpositive number, will not limit
   *       the number of nodes.
   */
  int max_visit;

  /**
   * The ground-truth permutation, only valid if @c learning is true.
   */
  Permutation truth;
  vector<int> true_cclist;

  double maxsofar;
  vector<int> bestseq;

  double negationoffset;

  // The transforms at various levels of the tree
  Transform::Left *Tgraph, *Tcorr;

private:
  QAPBnB(const QAPBnB &);
  const QAPBnB &operator = (const QAPBnB &);
};

class QAPbranch
{
public:
  QAPbranch(QAPBnB &owner, int m, int n);
  QAPbranch(QAPBnB &owner, const QAPSum &QAPs); 
  ~QAPbranch();


  double BranchAndBound(double maxsofar, const QAPSum::Weights &w);
 
  double GilmoreLawler();

  /**
   * Construct child node @c i
   */
  QAPbranch *makeBranch(int i) const;

  /**
   * Create child nodes
   */
  std::vector<QAPbranch *> makeChildren() const;

  /**
   * Calculate an upper bound on costs for solutions below this node.
   *
   * @note For leaves will return the objective value.
   */
  double bound(const QAPSum::Weights &w);

  bool isLeaf() const;
  int depth() const;
  int trueChild() const;

  QAPBnB &owner;
  const int m, n;

  std::vector<Fourier *> F;
  double *bounds;


  //Fourier reduceBound(const Fourier &original, const Fourier &unary, const Fourier &quadratic, double max, int ncomponents);

private:
  QAPbranch(const QAPbranch &);
  const QAPbranch &operator = (const QAPbranch &);
};


#endif // #ifndef __QAPbranch_HPP__
