#include <cassert>
#include <stdint.h>
#include <iostream>
#include <iomanip>
using std::cout; using std::endl;

#include <boost/shared_ptr.hpp>
#include <boost/shared_array.hpp>

#include "StochasticGradient.hpp"

using std::vector;

#ifndef MAX
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#endif

/**
 * A wrapper around QAP branch-and-bound nodes that does cleaner GC
 */
class Node
{
public:
  int m, n;

  Node(const QAPSum &qap, const Permutation &truth);
  ~Node();

  const double *getBounds();
  Node *child(int i);
  Node *trueSibling();
  Node *trueChild();

  bool isLeaf() const {return n == 1;}
  bool isTruth() const;

  int ndescendants() const;

  void print(std::ostream &) const;

private:

  Node *parent;
  int child_index;

  bool truth;
  boost::shared_ptr<QAPBnB> bnb;
  boost::shared_ptr<QAPbranch> branch;
  boost::shared_array<double> bounds;

  vector<Node> children;
  int true_child;

  Node(Node *parent, int i);
  //Node(const Node &);
  //const Node &operator = (const Node &);

  QAPbranch *getBranch();
  int getTrueChild();
};

std::ostream &operator << (std::ostream &stream, const Node &nd)
{
  nd.print(stream);
  return stream;
}

/**
 * Construct a root node for the tree.
 */
Node::Node(const QAPSum &qap, const Permutation &truth) :
  m(qap.m), n(qap.n),
  parent(NULL), child_index(-1), true_child(-1),
  bnb(new QAPBnB(qap, truth))
{
}

Node::Node(Node *init_parent, int i) :
  m(init_parent->m), n(init_parent->n - 1),
  parent(init_parent), child_index(i), true_child(-1)
{
}
Node::~Node() {}

const double *Node::getBounds()
{
  // Generate bounds
  if(bounds.get() == 0) {
    bounds = boost::shared_array<double>(new double[m]);
    
    getBranch()->bound(QAPSum::Weights(m));
    for(int i = 0; i < m; i++) {
      bounds[i] = branch->bounds[i];
    }

    if(!isTruth() || isLeaf()) { // Don't need QAPbranch anymore
      branch.reset();
    }
  }

  return bounds.get();
}

QAPbranch *Node::getBranch()
{
  if(branch.get() == 0) {
    // Generate branch as necessary
    if(parent == NULL) {
      branch = boost::shared_ptr<QAPbranch>(new QAPbranch(*bnb, bnb->QAPs));
    }
    else {
      QAPbranch *parent_branch = parent->getBranch();
      branch = boost::shared_ptr<QAPbranch>(parent_branch->makeBranch(child_index + 1));
    }

    if(!isLeaf()) {
      true_child = branch->trueChild() - 1;
    }
  }

  return branch.get();
}

int Node::getTrueChild()
{
  if(true_child == -1 && !isLeaf()) {
    getBranch();
  }

  return true_child;
}

Node *Node::child(int i)
{
  if(children.size() == 0) {
    children.reserve(n);
    for(int j = 0; j < n; j++) {
      children.push_back(Node(this, j));
    }
  }

  assert(i < children.size());
  return &children[i];
}

Node *Node::trueSibling(){
  return parent->child(parent->getTrueChild());
}

Node *Node::trueChild() {
  assert(!isLeaf());
  return child(getTrueChild());
}

/**
 * Whether this is the correct node.
 */
bool Node::isTruth() const
{
  assert(parent != NULL);
  return parent->getTrueChild() == child_index;
}

int Node::ndescendants() const
{
  return (n * (n+1)) / 2;
}

void Node::print(std::ostream &stream) const
{
  if(parent != NULL) {
    parent->print(stream);
    stream<<' ';
  }

  stream<<child_index;
}

/**
 * Select a random node out of all the BnB trees to reduce the objective on
 */
static Node *selectNode(std::vector<Node> &trees)
{
  // Select random training instance
  Node *nd = &trees[rand() % trees.size()];

  // Select random initial child
  int opt = rand() % 2;
  if(opt == 0) {
    nd = nd->trueChild();
  }
  else {
    nd = nd->child(rand() % nd->n);
  }

  while(!nd->isLeaf() && nd->isTruth()) {
    int opt = rand() % 3;
    if(opt == 0) {
      return nd;
    }
    else if(opt == 1) {
      nd = nd->trueChild();
    }
    else {
      nd = nd->child(rand() % nd->n);
    }
  }

  return nd;
}

static void iteration(
  std::vector<Node> &trees, QAPSum::Weights &w,
  double lambda, double eta)
{
  Node *nd = selectNode(trees);
  const double *bnds = nd->getBounds();

  const double *truebnds = nd->trueSibling()->getBounds();
  int m = nd->m;
  assert(m == (nd->trueSibling()->m));

  QAPSum::Weights grad(m);

  //cout<<QAPSum::Weights(m, bnds)<<" vs "<<QAPSum::Weights(m, truebnds)<<endl;
  //cout<<(*nd)<<endl;

  // Calculate this term of the objective
  double obj = 0;
  for(int i = 0; i < m; i++) {
    obj += w[i] * (bnds[i] - truebnds[i]);
  }

  // Calculate data term of objective
  if(obj + 1 > 0) {
    // Note that this path will also be followed for the true child
    // But the data components will be 0 anyway
    for(int i = 0;i < m;i++){
      grad[i] = (bnds[i] - truebnds[i]) + lambda * w[i];
    }
  }
  else {
    for(int i = 0; i < m; i++){
      grad[i] = lambda * w[i];
    }
  }

  // Update w
  for(int i = 0; i < m; i++){
    w[i] -= eta * grad[i];
  }

  // Project back onto positive
  for(int i = 0; i < m; i++) {
    if(w[i] < 0) {
      w[i] = 0;
    }
  }
}

void stochasticGradient(
  const std::vector<QAPSum> &train, const std::vector<Permutation> &truth,
  QAPSum::Weights &w, int niters, double lambda)
{
  // cout<<"niters "<<niters<<endl;
  double eta = 1;

  // Initialize BnB trees
  vector<Node> trees;
  trees.reserve(train.size());

  assert(train.size() == truth.size());
  for(int i = 0; i < train.size(); i++) {
    trees.push_back(Node(train[i], truth[i]));
  }
  cout<<"Starting stochastic gradient..."<<endl;

  // Iterate over nodes
  for(int iter = 0; iter < niters; iter++) {
    iteration(trees, w, lambda, eta);

    if(iter % 1024 == 0 && iter != 0) {
      cout<<"  iter "<<iter<<'/'<<niters;
      cout<<"  eta = "<<eta;
      cout<<"  w = "<<w;
      cout<<endl;
      //getchar();
    }

    eta *= (1 - 1e-3);
  }
 
  // Print out final weights
  cout<<"Final weights: "<<endl;
  int label_len = 0;
  for(int q = 0; q < train[0].m; q++) {
    label_len = MAX(label_len, train[0][q].label.size());
  }

  for(int q = 0; q < train[0].m; q++) {
    cout<<std::setw(3)<<q<<' '<<train[0][q].label;
    for(int i = train[0][q].label.size(); i <= label_len; i++) {
      cout<<' ';
    }
    cout<<std::setw(8)<<std::setprecision(4)<<w[q]<<endl;
  }
}
