#include <cassert>

#include "WTree.hpp"
#include "QAPSum.hpp"


WTree::WTree(const QAPSum &Q) :
  _parent(NULL), _j(0),
  _nchildren(Q.n), isdefault(true)
{
  // Create tree
  children = new WTree *[_nchildren];
  for(int i = 0; i < _nchildren; i++) {
    children[i] = NULL;
  }
// Create transforms
  Tgraph = new Transform::Left(Q.n, Q.n - 2);
  Tcorr = new Transform::Left(OuterProduct<Transform>(*Tgraph, *Tgraph));

   std:: vector<double> w (Q.m,1);
}

WTree::WTree(WTree *parent, int j, int nchildren, const vector<double> *initW) :
  _parent(parent), _j(j),
  _nchildren(nchildren),
  Tgraph(NULL), Tcorr(NULL)
{
  // Create tree
  children = new WTree *[nchildren];
  for(int i = 0; i < nchildren; i++) {
    children[i] = NULL;
  }

  // Initialize parameter
  if(initW == NULL) {
    w = NULL;
    isdefault = true;
  }
  else {
    std::vector<double> w (*initW);
    isdefault = false;
  }
}

WTree::~WTree()
{
  delete w;

  assert(children != NULL);

  for(int i = 0; i < _nchildren; i++) {
    delete children[i];
  }
  delete [] children;
  delete Tgraph;
  delete Tcorr;
  
}

WTree *WTree::child(int j)
{
  assert(children != NULL);
  assert(j >= 1 && j <= _nchildren);

  if(children[j-1] == NULL) {
    children[j-1] = new WTree(this, j, _nchildren - 1);
  }

  return children[j-1];
}

void WTree::setW(const vector<double> &newW)
{
  delete w;
  std::vector<double> w (newW);
  isdefault = false;
}

const vector<double> &WTree::getW()
{
  if(w == NULL) {
    // Derive default Fourier transform, filled with ones
    assert(_parent != NULL);

    const vector<double> &parentW = _parent->getW();
    const FourierT &T = *(parentW.T.descendant);
    
    // Fill with 1s
    w = new Fourier(T, 1.0);
  }

  return *w;
}

int WTree::nexpanded() const
{
  int n = 1;

  for(int i = 0; i < _nchildren; i++) {
    if(children[i] != NULL) {
      n += children[i]->nexpanded();
    }
  }

  return n;
}

const FourierT &WTree::getT()
{
  if(Tcorr == NULL) {
    assert(Tgraph == NULL);

    int n = nchildren();
    Tgraph = new Transform::Left(n, n - 2);
    Tcorr = new Transform::Left(OuterProduct<Transform>(*Tgraph, *Tgraph));
  }

  return Tcorr->top();
}

void WTree::normalizeW(int ncomponents, bool recursive)
{
  if(w == NULL) {
    // Default w is already normalized
    return;
  }

  if(ncomponents < 0) {
    ncomponents = w->size();
  }
  if(ncomponents == 0) {
    return;
  }

  // Find normalization factor
  double total = 0, nentries = 0;
  for(int k = 0; k < ncomponents; k++) {
    const FGmatrix &C = *((*w)[k]);

    nentries += C.Nrows * C.Ncols;
    for(int i = 0; i < C.Nrows; i++) {
      for(int j = 0; j < C.Ncols; j++) {
	total += C(i, j);
      }
    }
  }

  // Normalize
  assert(total != 0);
  for(int k = 0; k < ncomponents; k++) {
    FGmatrix &C = *((*w)[k]);

    for(int i = 0; i < C.Nrows; i++) {
      for(int j = 0; j < C.Ncols; j++) {
	C(i, j) *= nentries/total;
      }
    }
  }

  // Recurse
  if(recursive) {
    for(int i = 0; i < nchildren(); i++) {
      if(children[i] != NULL) {
	children[i]->normalizeW();
      }
    }
  }
}

////////////////////////// Input/Output ////////////////////
void WTree::write(ostream &stream) const
{
  // Output which children to expand
  stream<<nchildren()<<' ';


  int nexpanded = 0;
  for(int i = 0; i < nchildren(); i++) {
    if(children[i] != NULL) {
      nexpanded++;
    }
  }
  stream<<nexpanded<<' ';


  for(int i = 0; i < nchildren(); i++) {
    if(children[i] != NULL) {
      stream<<i<<' ';
    }
  }

  // Write w
  if(isDefault()) {
    stream<<"NULL"<<endl;
  }
  else {
    stream<<endl<<"w="<<endl;
    stream<<w->size()<<endl;

    for(int i = 0; i < w->size(); i++) {
      const Component &C = *(w->at(i));

      stream<<C.Nrows<<' '<<C.Ncols<<endl;
      stream<<C<<endl;
    }
  }
  stream<<endl;

  for(int i = 0; i < nchildren(); i++) {
    if(children[i] != NULL) {
      children[i]->write(stream);
    }
  }
}

void WTree::read(std::istream &stream)
{
  // Read # of children
  int read_nchildren = -1;
  stream>>read_nchildren;
  assert(read_nchildren == _nchildren);

  // Clear children
  for(int i = 0; i < nchildren(); i++) {
    delete children[i];
    children[i] = NULL;
  }

  // Read expanded children
  int nexpanded = -1;
  stream>>nexpanded;
  assert(nexpanded >= 0 && nexpanded <= _nchildren);

  std::vector<int> expanded(nexpanded);
  for(int i = 0; i < nexpanded; i++) {
    stream>>expanded[i];
  }

  // Read w
  std::string wtype;
  stream>>wtype;

  if(wtype == "NULL") {
    if(_parent != NULL) {
      delete w;
      w = NULL;
    }
  }
  else {
    assert(wtype == "w=");
    getW();
 

    int ncomponents;
    stream>>ncomponents;
    assert(ncomponents == w->size());

    for(int k = 0; k < ncomponents; k++) {
      Component &C = *(w->at(k));

      int rows, cols;
      stream>>rows>>cols;
      assert(rows == C.Nrows && cols == C.Ncols);

      for(int i = 0; i < rows; i++) {
	for(int j = 0; j < cols; j++) {
	  stream>>C(i, j);
	}
      }
    }

    //cout<<getW()<<endl;
  }

  // Read expanded children
  for(int i = 0; i < nexpanded; i++) {
    int j = expanded[i];

    child(j+1)->read(stream);
  }
}
