#ifndef __WTREE_HPP__
#define __WTREE_HPP__

#include <iostream>

#include "SnSFfourier.hpp"
#include "SnSFfourierT.hpp"
#include "SnSFtransform.hpp"

using namespace SnSF;

class QAPSum;

/**
 * A lazily expanded tree of QAP parameters.
 *
 * This class is built for having a unique Fourier parameter at each
 * node of a BnB tree.  Both the tree and the Fourier parameters are
 * lazily generated, with parameters always starting as all 1's.
 */
class WTree
{
public:
  /**
   * Create root node of a parameter tree.
   */
 
  WTree(const QAPSum &Q);

  virtual ~WTree();

  /**
   * Retrieve jth child of this node.
   *
   * @note Children are 1-indexed
   */
  WTree *child(int j);

  /**
   * Set new value for the parameter(vector of doubles) for this node.
   */
  void setW(const std::vector<double> &w);

  /**
   * Get parameter (vector of doubles) at this point in the tree.
   */
  const std::vector<double> &getW();

  /**
   * Multiply W by a constant to force the average element to be 1.
   *
   * @arg @c ncomponents Enforce this to be true for the first @c ncomponents
   *                     components of @c w.  Set to -1 for all components.
   * @arg @c recursive If @c true, also normalize all children.
   */
  void normalizeW(int ncomponents = -1, bool recursive = false);

  WTree *parent() {return _parent;}
  const WTree *parent() const {return _parent;}
  int j() const {return _j;}
  int nchildren() const {return _nchildren;}

  /**
   * Count number of descendants (and this node) which have been instantiated.
   */
  int nexpanded() const;

  /**
   * Whether this branch's w differs from the default all-ones
   */
  bool isDefault() const {return isdefault;}

  const FourierT &getT();

  /**
   * Read the contents of WTree from a file.
   */
  void read(std::istream &stream);

  /**
   * Write the WTree to a stream.
   */
  void write(ostream &stream) const;


protected:
  WTree *_parent;
  int n, _j;

  int _nchildren;
  WTree **children;

  bool isdefault;
  std::vector<double> *w;

  WTree(WTree *parent, int j, int nchildren, const Fourier *initW = NULL);

  // Stuff to generate Fourier transforms from QAP
  SnSF::Transform::Left *Tgraph, *Tcorr;

private:
  WTree(const WTree &copy);
  const WTree &operator = (const WTree &assign);
};

#endif // #ifndef __WTREE_HPP__
