#include <cmath>
#include <algorithm>
#include <cassert>

#include "learn.hpp"
#include "AssignmentMatrix.hpp"

std::vector<int> readNodeList(const char *path)
{
  cout<<path<<endl;

  std::vector<int> nodes;
  std::ifstream file(path, std::ios::in);
  
  while(!file.eof()) {
    int n = -1;
    file>>n;
    if(n == -1) {
      break;
    }

    cout<<n<<' ';
    nodes.push_back(n);
  }
  cout<<endl;

  return nodes;
}

inline static double square(double x) {return x*x;}

/**
 * Load the shape context features for house @c i
 */
static Graph makePhi(Fmatrix A, Fmatrix B)
{
  assert(A.Ncols == B.Ncols);
  assert(A.Nrows == B.Nrows);

  Graph Phi(A.Nrows);
  for(int i = 0; i < A.Nrows; i++) {
    for(int j = 0; j < B.Nrows; j++) {
      // Phi(i, j) = distance(A(i, :), B(j, :))

      double dist = 0;
      for(int k = 0; k < A.Ncols; k++) {
	dist += square(A(i, k) - B(j, k));
      }
      dist = sqrt(dist);

      Phi(i, j) = dist;
    }
  }

  return Phi;
}

list<QAP> loadFileList(const char *path)
{
  list<QAP> trainset;
  std::ifstream listfile(path, std::ios::in);

  // Read in files
  while(!listfile.eof()) {
    string ptpath, scfpath, imgpath, adjpath;

    listfile>>ptpath;
    if(listfile.eof()) {
      break;
    }

    listfile>>scfpath;
    listfile>>imgpath;
    listfile>>adjpath;

    //cout<<scfpath<<' '<<adjpath<<' ';

    Fmatrix scf1(scfpath);
    Graph G1(adjpath);
  
    listfile>>ptpath;
    listfile>>scfpath;
    listfile>>imgpath;
    listfile>>adjpath;

    //cout<<scfpath<<' '<<adjpath<<endl;

    Fmatrix scf2(scfpath);
    Graph G2(adjpath);
    
    Graph k1 = makePhi(scf1, scf1);
    Graph k2 = makePhi(scf2, scf2);
   
    QAP qap(G1, G2, k1, k2);
      
 trainset.push_back(qap);
  }

  return trainset;
}

template<class T>
static std::list<T> shuffleList(const std::list<T> &list)
{
  typedef typename std::list<T>::const_iterator iterT;

  // Create list of iterator
  std::vector<iterT> iters(list.size());  
  {
    int i = 0;
    iterT iter = list.begin();
    for(; iter != list.end(); i++, iter++) {
      iters[i] = iter;
    }
  }

  // Shuffle the iterators
  random_shuffle(iters.begin(), iters.end());

  std::list<T> shuffled;
  for(int i = 0; i < iters.size(); i++) {
    shuffled.push_back(*iters[i]);
  }

  return shuffled;
}

void train(const std::list<QAP> &trainset, const std::vector<Permutation> &truth,
	   int nepochs, WTree *w)
{

  for(int epoch = 0; epoch < nepochs; epoch++) {
    std::list<QAP> epochset = shuffleList(trainset);

    int i = 0;
    for(std::list<QAP>::iterator iter = epochset.begin();
	iter != epochset.end() && i < truth.size();
	iter++, i++) {

      iter->train(truth[i], w);
    }
  }
}

std::vector<Permutation> test(const std::list<QAP> &QAPs, WTree *w)
{
  std::vector<Permutation> predict(QAPs.size());

  int i = 0;
  std::list<QAP>::const_iterator Q = QAPs.begin();
  for(; Q != QAPs.end(); i++, Q++) {
    assert(i < QAPs.size());

    predict[i] = Q->solve(w);
  }

  return predict;
}

double loss(const std::vector<Permutation> &truth, const std::vector<Permutation> &predict)
{
  assert(truth.size() == predict.size());

  double wrong = 0, total = 0;

  for(int i = 0; i < truth.size(); i++) {
    assert(truth[i].n == predict[i].n);

    total += (double)(truth[i].n);
    for(int j = 0; j < truth[i].n; j++) {
      //cout<<i<<' '<<j<<' '<<predict[i][j]<<' '<<truth[i][j]<<endl;

      if((predict[i])[j] != (truth[i])[j]) {
	wrong += 1.0;
      }
    }
  }

  cout<<"wrong "<<wrong<<endl;
  cout<<"total "<<total<<endl;
  return wrong/total;
}

std::vector<Permutation> munkres(const std::list<QAP> &QAPs)
{
  std::vector<Permutation> predict(QAPs.size());
  /*
  int i = 0;
  std::list<QAP>::const_iterator Q = QAPs.begin();
  for(; Q != QAPs.end(); i++, Q++) {
    assert(i < QAPs.size());
    int n = Q->n;
  */
    /*
    Permutation::Random rand(n);
    cout<<rand<<endl;

    cout<<"Q->phi = "<<endl<<Q->phi<<endl;
    Graph Phi(n);
    for(int a = 0; a < n; a++) {
      for(int b = 0; b < n; b++) {
	Phi(a, rand[b]-1) = Q->phi(a, b);
      }
    }
    cout<<"Phi = "<<endl<<Phi<<endl;
    //*/

    // Solve Munkres problem
    /*AssignmentMatrix A(Q->phi);
    A.KuhnMunkres();

    // Predict
    predict[i] = Permutation(n);
    for(int j = 0; j < n; j++) {
      predict[i][j] = A.assignedx[j]+1;
    }
  }
    */

  return predict;
}
