#include <cstring>
#include <iostream>
#include <sysexits.h>
#include <cstdlib>
#include <cassert>

using std::cout; using std::endl;

#include "Fold.hpp"
#include "ImageDatasets.hpp"
#include "ImageSequence.hpp"
#include "Learners.hpp"

static const char * const USAGE = "\
Usage: seqfold [OPTS]\n\
\n\
Arguments:\n\
  -d, --dataset: Which image sequence to run on of house, hotel, skew, etc.  [Required]\n\
  -b, --baseline: Baseline distance. [Required]\n\
  -f, --folds: Number of folds. [Required]\n\
  -a, --algorithm: The algorithm to use for learning. [Required]\n\
\n\
  --fold-input: File from which to read folds.\n\
  --fold-output: File to output folds.\n\
  -R, --results: File to output detailed results.\n\
  --max-nodes: Limit the BnB to visiting a given number of nodes.\n\
  --threads: The number of threads to execute.  Will distribute folds between threads.\n\
\n\
  -h, --help: Print this message and exit.\n\
\n";

static bool argmatch(const char *test, char brief, const char *full);

static bool argpresent(int argc, char **argv, char brief, const char *full);
static const char *argvalue(int argc, char **argv, char brief, const char *full);

static int argvalueint(int argc, char **argv, char brief, const char *full, bool required, int def);

int main(int argc, char **argv)
{
  // Read arguments
  if(argpresent(argc, argv, 'h', "help")) {
    cout<<"Help requested"<<endl;
    cout<<USAGE;
    return EX_OK;
  }

  // Random seed
  srand(0);

  // Load dataset
  const ImageDataset *dataset = getDataset(argvalue(argc, argv, 'd', "dataset"));
  if(dataset == NULL) {
    cout<<"Valid dataset not specified"<<endl<<USAGE;
    return EX_USAGE;
  }

  // Set up learner
  int baseline = argvalueint(argc, argv, 'b', "baseline", true, 0);
  int nfolds = argvalueint(argc, argv, 'f', "folds", true, 0);

  Learner *learner = NULL;
  const char *learner_name = argvalue(argc, argv, 'a', "algorithm");
  int max_nodes = argvalueint(argc, argv, '\0', "max-nodes", false, -1);
  int nthreads = argvalueint(argc, argv, '\0', "threads", false, 1);
  if(learner_name == NULL) {
    cout<<"Did not specify which algorithm"<<endl<<USAGE;
    return EX_USAGE;
  }
  if(strcmp(learner_name, "bnb") == 0) {
    learner = new BnBLearner(false, max_nodes);
  }
  else if(strcmp(learner_name, "greedy") == 0) {
    learner = new BnBLearner(true);
  }
  else if(strcmp(learner_name, "nolearn") == 0) {
    learner = new NoLearner(false, max_nodes);
  }
  else if(strcmp(learner_name, "nolearn-greedy") == 0) {
    learner = new NoLearner(true);
  }
  if(learner == NULL) {
    cout<<"Did not specify valid algorithm"<<endl<<USAGE;
    return EX_USAGE;
  }

  const char *fold_input_path = argvalue(argc, argv, '\0', "fold-input");
  const char *fold_output_path = argvalue(argc, argv, '\0', "fold-output");
  const char *results_path = argvalue(argc, argv, 'R', "results");

  cout<<"Running on "<<dataset->name<<" dataset with baseline "<<baseline<<" on "<<nfolds<<" folds"<<endl;

  // Set up dataset
  std::vector<QAPSum> Qs = imageSequence(dataset->n, baseline, dataset->adj_format, dataset->scf_format, dataset->dis_format);

  // Construct folds
  std::vector<Fold> folds;
  if(fold_input_path == NULL) {
    std::ofstream fold_output;
    std::ostream *fold_output_stream = NULL;
    if(fold_output_path != NULL) {
      cout<<"Outputting folds to "<<fold_output_path<<endl;
      fold_output.open(fold_output_path, std::ios::out | std::ios::trunc);
      fold_output_stream = &fold_output;
    }

    folds = Fold::nfold(Qs, nfolds, fold_output_stream);

    if(fold_output_path != NULL) {
      fold_output.close();
    }
  }
  else {
    std::ifstream fold_input(fold_input_path);
    if(fold_input.fail()) {
      std::cerr<<"Failed to open fold input file \""<<fold_input_path<<'"'<<endl;
      return EX_NOINPUT;
    }
    folds = Fold::from_file(Qs, nfolds, fold_input);
  }

  // Run the learning
  std::vector<std::vector<TestResult> > results = Fold::run(folds, *learner, nthreads);

  // Write detailed output
  if(results_path != NULL) {
    std::ofstream results_out(results_path, std::ios::out | std::ios::trunc);
    if(results_out.fail()) {
      std::cerr<<"Error: cannot open detailed results file \""<<results_path<<'"'<<endl;
      return EX_CANTCREAT;
    }

    for(int i = 0; i < nfolds; i++) {
      // Output fold summary
      
      results_out<<"Fold "<<(i + 1)<<endl;
      results_out<<"  accuracy: "<<TestResult::accuracy(results[i].begin(), results[i].end())<<endl;
      results_out<<"  nvisited: "<<TestResult::averageVisited(results[i].begin(), results[i].end())<<endl;
      results_out<<endl;

      assert(folds[i].test_ind.size() == results[i].size());


      // Output test results
      for(int j = 0; j < results[i].size(); j++) {
	int ind = folds[i].test_ind[j];
	results_out<<"  ("<<(ind + 1)<<", "<<(ind + baseline + 1)<<")  "<<results[i][j]<<endl;
      }

      results_out<<endl;
    }

    // Output total summary
    vector<TestResult> all = Fold::mergeResults(results);
    results_out<<"Overall"<<endl;
    results_out<<"  accuracy: "<<TestResult::accuracy(all.begin(), all.end())<<endl;
    results_out<<"  nvisited: "<<TestResult::averageVisited(all.begin(), all.end())<<endl;


    results_out.close();
  }

  // Cleanup
  delete learner;

  return EX_OK;
};

static bool argmatch(const char *test, char brief, const char *full)
{
  if(brief != '\0') {
    if(test[0] == '-' && test[1] == brief && test[2] == '\0') {
      return true;
    }
  }
  if(full != NULL) {
    if(test[0] == '-' && test[1] == '-' && strcmp(test + 2, full) == 0) {
      return true;
    }
  }
  return false;
}

static bool argpresent(int argc, char **argv, char brief, const char *full)
{
  for(int i = 1; i < argc; i++) {
    if(argmatch(argv[i], brief, full)) {
      return true;
    }
  }
  return false;
}

static const char *argvalue(int argc, char **argv, char brief, const char *full)
{
  for(int i = 1; i < argc-1; i++) {
    if(argmatch(argv[i], brief, full)) {
      return argv[i+1];
    }
  }
  return NULL;
}

static int argvalueint(int argc, char **argv, char brief, const char *full, bool required, int def)
{
  const char *str = argvalue(argc, argv, brief, full);

  if(str == NULL) {
    if(required) {
      cout<<"Missing required argument --"<<full<<endl<<USAGE;
      exit(EX_USAGE);
    }

    return def;
  }

  return atoi(str);
}
