#include "SnSFfourierT.hpp"

#include "SnSFcomponentT.hpp"
#include "SnSFtransform.hpp"

SnSF::FourierT::FourierT(const int _n, const int _order):
  n(_n), order(_order),  ancestor(NULL), descendant(NULL)
{
}



SnSF::FourierT::FourierT(const OuterProduct<FourierT> op):
  n(op.first.n), ancestor(NULL), descendant(NULL)
{
  const FourierT& first=op.first;
  const FourierT& second=op.second;
  for(int i = 0; i < first.size(); i++) {
    push_back(new ComponentT(OuterProduct<ComponentT>(*first[i], *second[i])));
  }
}


SnSF::FourierT::FourierT(const InnerProduct<FourierT> op):
  n(op.first.n), ancestor(NULL), descendant(NULL)
{
  const FourierT& first=op.first;
  const FourierT& second=op.second;
  for(int i = 0; i < first.size() && i < second.size(); i++) {
    push_back(new ComponentT(InnerProduct<ComponentT>(*first[i], *second[i])));
  }
}

int SnSF::FourierT::operator[] (const PartitionB& shape) const
{
  for(int i = 0; i < size(); i++) {
    if(at(i)->shape == shape) {
      return i;
    }
  }
  return -1;
}

SnSF::FourierT *SnSF::FourierT::makeAncestorLeft(const PartitionB* limit)
{
  FourierT *R = new FourierT(n+1);
  int Nexisting = size();

  for(int compi = 0; compi < Nexisting; compi++) {
    PartitionB p = at(compi)->shape;
    if(p.size() == 1 || p[p.size()-1] < p[p.size()-2]) {
      p[p.size()-1]++;
      if(limit == NULL || p.limitedby(*limit)) {
	R->push_back(new ComponentT(p,this));
      }
      p[p.size()-1]--;
    }
    p.push_back(1);
    if(limit == NULL || p.limitedby(*limit)) {
      R->push_back(new ComponentT(p,this));
    }
  }

  ancestor = R;
  R->descendant = this;
  R->order = order * (n+1);

  return R;
}


SnSF::FourierT *SnSF::FourierT::makeDescendantLeft()
{
  FourierT *R = new FourierT(n-1);

  for(int compi = 0; compi < size(); compi++){
    ComponentT& comp = *at(compi);
    int row = 0, col = 0;

    while(row < comp.Nrows){
      int pos=comp.rowlabels[row].row[n];
      PartitionB subShape(comp.shape);
      subShape[pos-1]--;
      if(subShape[pos-1] == 0) {
	subShape.pop_back();
      }

      int subi=(*R)[subShape];
      if(subi==-1){
	ComponentT& sub= *new ComponentT(subShape);
	if(comp.dyadic) sub.dyadic=1;
	while(row<comp.Nrows && comp.rowlabels[row].row[n]==pos){
	  Tableau t(comp.rowlabels[row++]);
	  t.row.pop_back();
	  t.col.pop_back();
	  t.shape[pos-1]--;
	  if(t.shape[pos-1]==0) t.shape.pop_back();
	  sub.rowlabels.push_back(t);
	}
	while(col<comp.Ncols && comp.columnlabels[col].row[n]==pos){
	  Tableau t(comp.columnlabels[col++]);
	  t.row.pop_back();
	  t.col.pop_back();
	  t.shape[pos-1]--;
	  if(t.shape[pos-1]==0) t.shape.pop_back();
	  sub.columnlabels.push_back(t);
	}
	sub.Nrows=sub.rowlabels.size();
	sub.Ncols=sub.columnlabels.size();
	sub.computeYORleft();
	sub.pseudo=0;
	R->push_back(&sub);
	subi=R->size()-1;
      }else{
	row+=(*R)[subi]->Nrows;
	col+=(*R)[subi]->Ncols;
	(*R)[subi]->dyadic=0;
      }
      comp.sub.push_back(subi);
    }
  }

  descendant = R;
  R->ancestor = this;

  return R;
}


string SnSF::FourierT::str() const
{
  ostringstream result;
  result<<size()<< " components"<<endl;
  for(int i=0; i<size(); i++) {
    result<<at(i)->str();
  }
  return result.str();
}

int SnSF::FourierT::readfrom(ifstream& ifs)
{
  int t; 
  ifs.read(reinterpret_cast<char*>(&n),sizeof(int));
  ifs.read(reinterpret_cast<char*>(&order),sizeof(double));
  ancestor = descendant = NULL;
  ifs.read(reinterpret_cast<char*>(&t),sizeof(int));  
  clear();
  for(int i=0; i<t; i++){
    PartitionB p(1,NULL);
    ComponentT* compT=new ComponentT(p);
    compT->readfrom(ifs);
    push_back(compT);
  }
}



int SnSF::FourierT::writeto(ofstream& ofs) const
{
  int t;
  ofs.write(reinterpret_cast<const char*>(&n),sizeof(int));
  ofs.write(reinterpret_cast<const char*>(&order),sizeof(double));
  t=size(); ofs.write(reinterpret_cast<char*>(&t),sizeof(int));
  for(int i=0; i<t; i++) at(i)->writeto(ofs);
}

