import cairo
import sys, os, os.path

USAGE = """python plotpairs.py output image1 nodes1 image2 nodes2 permutation node_set

Plot a mapping between a pair of images according to a permutation of node
indices.

Arguments:
  output: Path to write an output PNG or PDF file.
  image1,image2: Images on which to plot the points.
  nodes1,nodes2: Positions of nodes in the images, i.e. 'house/houses/house1'
  permutation: File containing a permutation.
  node_set: File specifying a subset of nodes.  (Optional)
"""

def loadPoints(points_path, nodes = None):
    p = []
    for line in file(points_path):
        line = line.strip()
        if line == '':
            continue

        line = [x for x in line.split() if x != '']
        line = float(line[0]),float(line[1])

        p.append(line)

    if nodes != None:
        p = [p[i-1] for i in nodes]

    return p

def loadPermutation(perm_path):
    s = file(perm_path).read()
    s = s.strip()
    s = map(int, s.split(' '))
    return s

def plotPair(outpath, img1, p1, img2, p2, permutation):
    # Load the two images
    img1 = cairo.ImageSurface.create_from_png(img1)
    img2 = cairo.ImageSurface.create_from_png(img2)

    width = img1.get_width() + img2.get_width()
    assert(img1.get_height() == img2.get_height())
    height = img1.get_height()

    # Allocate surface
    ext = os.path.splitext(outpath)[1].lower()
    if ext == '.png':
        surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
    elif ext == '.pdf':
        surf = cairo.PDFSurface(outpath, width, height)
        
    cr = cairo.Context(surf)

    # Composite the two images
    cr.set_source_surface(img1, 0, 0)
    cr.paint()

    cr.translate(img1.get_width(), 0)
    cr.set_source_surface(img2, 0, 0)
    cr.paint()
    cr.translate(-img1.get_width(), 0)

    # Draw correspondence
    cr.set_line_width( max(cr.device_to_user_distance(2, 2)) )
    for src,dst in enumerate(permutation):
        dst = dst-1

        cr.set_source_rgb(0, 1, 0)

        #if src == dst:
        #    cr.set_source_rgb(0, 0, 1)
        #else:
        #    cr.set_source_rgb(1, 0, 0)

        src = p1[src]
        dst = p2[dst]

        cr.move_to(src[0], src[1])
        cr.line_to(dst[0] + img1.get_width(), dst[1])
        cr.stroke()

    # Write output
    if ext == '.png':
        surf.write_to_png(outpath)
    surf.finish()

def getPair(which, offset, n):
    path = 'pairs/houses%d_%s.txt' % (offset, which)

    # Get index for each line of the input file
    houses = []
    for line in file(path, 'r'):
        line = line.strip()
        if line == '':
            continue

        line = [x for x in line.split(' ') if x != '']
        line = line[0][len('data/house/houses/house'):]
        houses.append(int(line))
    assert(len(houses) % 2 == 0)
    #print houses

    # Group pairs of lines
    pairs = [(houses[2*i],houses[2*i+1]) for i in range(len(houses)/2)]
    #print pairs

    return (pairs[n][0], pairs[n][1])

# Interpret arguments
if '-h' in sys.argv or '--help' in sys.argv:
    print USAGE
    exit(0)
if len(sys.argv) not in [7,8]:
    print 'Wrong number of arguments.\nUSAGE:'
    print USAGE

if len(sys.argv) > 7:
    node_set = loadPermutation(sys.argv[7])
else:
    node_set = None

output = sys.argv[1]
img1 = sys.argv[2]
points1 = loadPoints(sys.argv[3], node_set)
img2 = sys.argv[4]
points2 = loadPoints(sys.argv[5], node_set)
permutation = loadPermutation(sys.argv[6])

if node_set != None:
    assert(len(permutation) == len(node_set))

# Plot the pair
plotPair(output, img1, points1, img2, points2, permutation)
print 'Plotted to %s' % output
