#include "Graph.hpp"

#include "SnSF.hpp"
#include "SnSFfunction.hpp"
#include "SnSFfourier.hpp"

#include <stdarg.h>
#include <cstring>
#include <algorithm>

Graph::Graph(const int n) : 
  Fmatrix(n,n)
{}

Graph::Graph(const int n, const FIELD val): Fmatrix(n,n,val){}



Graph::Graph(const char* s1, ...):Fmatrix(strlen(s1),strlen(s1)){

  va_list params;
  const char* arg=s1;
  va_start(params,s1);
  for(int i=0; i<Nrows; i++){
    for(int j=0; j<strlen(arg); j++)
      if (arg[j]=='*') at(i,j)=1; else at(i,j)=0;
    arg=va_arg(params, const char*);
  }
  va_end(params);
}


Graph::Graph(const Graph &o) :
  Fmatrix(o)
{}

Graph::Graph(const Fmatrix &F) :
  Fmatrix(F)
{}

const Graph &Graph::operator = (const Graph &G)
{
  Fmatrix::operator =(G);
  return *this;
}

Graph::~Graph()
{
}

SnSF::Function* Graph::function(const SnSF::Transform &_transfm) const
{
  SnSF::Function &R = *new SnSF::Function(_transfm);
  const int n = Nrows;

  if(R.T.order == n){
    for(int i = 1; i <= R.T.order; i++) {
      R.f[i] = i;
    }
  }
  else{
    for(int i=1; i<=n; i++)
      for(int j=1; j<=n; j++)
	if(i!=j) {
	  assert((n-i)*(n-1)+(n-1-(j-(i<j))) < R.T.order);
	  R.f[(n-i)*(n-1)+(n-1-(j-(i<j)))]=at(i-1,j-1);
	}
  }

  return &R;
}

SnSF::Fourier *Graph::Fourier(const SnSF::Transform &_transfm) const
{
  SnSF::Function *f = function(_transfm); 
  SnSF::Fourier *F = new SnSF::Fourier(*f);
  delete f;
  
  return F;
}


Graph::Cycle::Cycle(const int n, int m):
  Graph(n,0)
{
  if(m==0) m=n;
  for(int i=0; i<m; i++){
    const int j=(i+1)%m;
    at(n-1-i,n-1-j)=1;
    at(n-1-j,n-1-i)=1;
  }
}



Graph::Circuit::Circuit(const int n, int m):
  Graph(n,0)
{
  if(m==0) m=n;
  for(int i=0; i<m; i++){
    const int j=(i+1)%m;
    at(n-1-i,n-1-j)=1;
  }
}



Graph::Chain::Chain(const int n, int m):
  Graph(n,0)
{
  if(m==0) m=n;
  for(int i=0; i<m-1; i++){
    const int j=i+1;
    at(n-1-i,n-1-j)=1;
    at(n-1-j,n-1-i)=1;
  }
}



Graph::RandomSimple::RandomSimple(const int n, const double v):
  Graph(n)
{
  for(int i=0; i<Ncols; i++){
    for(int j=0; j<i; j++)
      if(((double) rand())/((double)RAND_MAX+1)>v){at(i,j)=1; at(j,i)=1;}
      else{at(i,j)=0; at(j,i)=0;}
    at(i,i)=0;
  }
}



Graph::RandomBinary::RandomBinary(const int n, const double v):
  Graph(n)
{
  for(int i=0; i<Ncols; i++){
    for(int j=0; j<Ncols; j++)
      if(((double) rand())/((double)RAND_MAX+1)>v){at(i,j)=1;}
      else{at(i,j)=0;}
    at(i,i)=0;
  }
}



Graph::RandomWeighted::RandomWeighted(const int n, const double thresh):
  Graph(n)
{
  for(int i=0; i<Ncols; i++)
    for(int j=0; j<Ncols; j++)
      if(i==j || ((double) rand())/((double)RAND_MAX+1)<thresh) at(i,j)=0;
      else at(i,j)=((double) rand())/((double)RAND_MAX+1);
}

Graph::RandomNedges::RandomNedges(int n, int nedges) :
  Graph(n, 0)
{
  assert(nedges < (n * (n-1)) / 2);

  // Generate list of all possible edges
  vector<pair<int, int> > edges((n * (n-1)) / 2);
  int k = 0;
  for(int i = 0; i < n; i++) {
    for(int j = i+1; j < n; j++) {
      edges[k].first = i;
      edges[k].second = j;
      k++;
    }
  }

  // Shuffle
  random_shuffle(edges.begin(), edges.end());

  // Grab first nedges
  for(int k = 0; k < nedges; k++) {
    int i = edges[k].first;
    int j = edges[k].second;

    (*this)(i, j) = (*this)(j, i) = 1;
  }
}

Graph Graph::permute(const Permutation &p) const
{
  Graph t(Nrows); 

  for(int i=0; i<Nrows; i++) {
    for(int j=0; j<Ncols; j++) {
      t(p[i]-1, p[j]-1) = at(i, j);
    }
  }

  return t;
}

Graph Graph::pseudoNegate()
{
  Graph neg(Nrows);

  assert(Nrows == Ncols && neg.Nrows == neg.Ncols);

  double max = 0;
  for(int i = 0; i < Nrows*Ncols; i++) {
    if(a[i] > max) {
      max = a[i];
    }
  }

  for(int i = 0; i < Nrows*Ncols; i++) {
    neg.a[i] = max - a[i];
  }

  return neg;
}

Graph Graph::subgraph(const std::vector<int> &nodes) const
{
  return Graph(matrixMinor(nodes, nodes));
}

std::ostream &operator << (std::ostream &s, const Graph &G)
{
  s<<G.Nrows<<'x'<<G.Ncols<<endl;

  for(int i = 0; i < G.Nrows; i++) {
    for(int j = 0; j < G.Ncols; j++) {
      s<<G(i,j)<<" ";
    }
    s<<endl;
  }

  return s;
}
