#ifndef _FGmatrix
#define _FGmatrix 

#include "base.h"
#include <string>
#include <vector>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <cassert>
#include "gsl/gsl_rng.h"
#include "gsl/gsl_randist.h"
#include <gsl/gsl_linalg.h>
#include "ExpressionTemplates.hpp"

//using namespace Eigen;
using namespace std;

class FGmatrix{
public:

  class Random;

  FGmatrix(const FGmatrix& o);
  FGmatrix(const bool _dyadic, const int _Nrows, const int _Ncols);
  FGmatrix(const bool _dyadic, const int _Nrows, const int _Ncols, const FIELD val);
  FGmatrix(const bool _dyadic, const int _Nrows, const int _Ncols, const Identity& dummy);
  FGmatrix(const Sum<FGmatrix, FGmatrix>& args);
  FGmatrix(const Product<FGmatrix, FGmatrix>& args);
  FGmatrix(const Product<FGmatrix, Transpose<FGmatrix> >& args);
  FGmatrix(const Product<Transpose<FGmatrix>, FGmatrix>& args);
  FGmatrix(const Product<Transpose<FGmatrix>, Transpose<FGmatrix> >& args);
  FGmatrix(const Product<FGmatrix, Inverse<FGmatrix> >& args);
  FGmatrix(const Product<Inverse<FGmatrix>, FGmatrix> & args);
  FGmatrix(const Product<FGmatrix, FIELD >& args);
  FGmatrix(const HadamardProduct<FGmatrix, FGmatrix> &args);
  FGmatrix(const HadamardDivision<FGmatrix, FGmatrix> &args);
  FGmatrix(const string filename);
  FGmatrix(ifstream& ifs, const int _Nrows, const int _Ncols, const bool binary=0);
  virtual ~FGmatrix(){delete [] a; if(dyadic) delete [] b;}

  const FGmatrix& operator  = (const FGmatrix& o);
  const FGmatrix& operator += (const FGmatrix& o);
  const FGmatrix& operator -= (const FGmatrix& o);
  const FGmatrix& operator *= (const FIELD o);
  
  Transpose<FGmatrix> operator~() const {return Transpose<FGmatrix>(*this);}


  FIELD& operator() (const int i, const int j) {
    assert(i < Nrows && j < Ncols);
    assert(!dyadic);

    return a[i*Ncols+j];
  }

  FIELD operator() (const int i, const int j) const {
    assert(i < Nrows && j < Ncols);

    if(dyadic) return a[i]*b[j];
    else return a[i*Ncols+j];
  } 

  FIELD& at(const int i, const int j){
    assert(i < Nrows && j < Ncols);
    assert(!dyadic);

    return a[i*Ncols+j];
  }

  FIELD at(const int i, const int j) const {
    assert(i < Nrows && j < Ncols);

    if(dyadic) return a[i]*b[j];
    else return a[i*Ncols+j];
  } 

  FIELD val(const int i, const int j) const {
    assert(i < Nrows && j < Ncols);

    if(dyadic) return a[i]*b[j];
    else return a[i*Ncols+j];
  } 

  /**
   * Calculate Frobenius norm of matrix.
   */
  FIELD norm2() const;
  FIELD inp(const FGmatrix& o) const;
  FIELD trace() const;

  /**
   * Whether any entry in the matrix is NaN.
   */
  bool isNaN() const;

  /**
   * Calculate singular values of the matrix.
   *
   * @see SVD
   */
  std::vector<FIELD> singularValues() const;

  /**
   * Calculate singular value decomposition.
   *
   * Returns orthogonal matrices U,V and diagonal matrix S such that
   * this matrix is equal to
   *
   * @[
   *    U S V^T
   * @]
   */
  void SVD(FGmatrix &S, FGmatrix &U, FGmatrix &V) const;
  
  /**
   * Find the pseudoinverse of the matrix.
   *
   * @arg @c tol The tolerance below which we decide an eigenvalue of the
   *             matrix is 0.
   */
  FGmatrix pseudoInverse(double tol = 1e-6) const;

  /**
   * Calculate trace norm.
   *
   * Also known as "nuclear" norm.  It is equal to the sum of the singular
   * values.
   */
  double tracenorm() const;

  string str() const;
  int writeto(ofstream& ofs, const bool binary=0);
  
  /**
   * Calculate QR decomposition of the matrix.
   *
   * @return A pair, with $Q$ in @c first and $R$ in @c second. 
   */
  std::pair<FGmatrix, FGmatrix> QR() const;

  /**
   * Solve an overdetermined or undetermined system A x = b
   *
   * @arg @c b Input RHS m x nrhs matrix
   * @return Solution n x nrhs matrix
   * @see QR
   */
  FGmatrix solveQR(const FGmatrix &b, double tol = 1e-6) const;

  /**
   * Solve an overdetermined or undetermined system A x = b using QR decomp.
   *
   * @arg @c b Input RHS m x nrhs matrix
   * @return Solution n x nrhs matrix
   * @see QR
   * @note Currently requires full-rank R
   */
  static FGmatrix solveQR(const FGmatrix &Q, const FGmatrix &R, const FGmatrix &b, double rcond = 1e-6);

  /**
   * Calculate transpose.
   */
  FGmatrix transpose() const;

  int Nrows;
  int Ncols;

  bool dyadic;
  FIELD* a;
  FIELD* b;

};

class FGmatrix::Random: public FGmatrix{
public:
  Random(const bool _dyadic, const int _Nrows, const int _Ncols);
};

ostream& operator<<(ostream& stream, const FGmatrix& o);

// Binary operators
FGmatrix operator + (const FGmatrix &A, const FGmatrix &B);
FGmatrix operator - (const FGmatrix &minuend, const FGmatrix &subtrahend);

inline Product<FGmatrix, FGmatrix> operator * (const FGmatrix &A, const FGmatrix &B)
{
  return Product<FGmatrix, FGmatrix>(A, B);
}

inline Product<Transpose<FGmatrix>, FGmatrix> operator * (const Transpose<FGmatrix> &A, const FGmatrix &B)
{
  return Product<Transpose<FGmatrix>, FGmatrix>(A, B);
}

inline Product<FGmatrix, Transpose<FGmatrix> > operator * (const FGmatrix &A, const Transpose<FGmatrix> &B)
{
  return Product<FGmatrix, Transpose<FGmatrix> >(A, B);
}

inline Product<Transpose<FGmatrix>, Transpose<FGmatrix> > operator * (const Transpose<FGmatrix> &A, const Transpose<FGmatrix> &B)
{
  return Product<Transpose<FGmatrix>, Transpose<FGmatrix> >(A, B);
}

#endif


