#include "Matlab.hpp"
#include "Format.hpp"

#include <cstdarg>

using namespace matlab;

FGmatrix matlab::FGmatrixFromMx(const mxArray *mx)
{
  assert(mxGetNumberOfDimensions(mx) == 2);
  size_t m = mxGetM(mx), n = mxGetN(mx);
  double *p = mxGetPr(mx);

  // Retrieve values
  FGmatrix f(false, m, n);
  for(size_t i = 0; i < m; i++) {
    for(size_t j = 0; j < n; j++) {
      f(i,j) = p[i + m*j];
    }
  }

  return f;
}

Fmatrix matlab::FmatrixFromMx(const mxArray *mx)
{
  assert(mxGetNumberOfDimensions(mx) == 2);
  size_t m = mxGetM(mx), n = mxGetN(mx);
  double *p = mxGetPr(mx);

  // Retrieve values
  Fmatrix f(m, n);
  for(size_t i = 0; i < m; i++) {
    for(size_t j = 0; j < n; j++) {
      f(i,j) = p[i + m*j];
    }
  }

  return f;
}

mxArray *matlab::FGmatrixToMx(const FGmatrix &f)
{
  // Allocate matlab matrix
  mwSize dims[2] = {f.Nrows, f.Ncols};
  mxArray *m = mxCreateNumericArray(2, dims, mxDOUBLE_CLASS, mxREAL);
  double *p = mxGetPr(m);

  // Fill
  for(int i = 0; i < f.Nrows; i++) {
    for(int j = 0; j < f.Ncols; j++) {
      p[i + f.Nrows*j] = f(i, j);
    }
  }

  return m;
}

mxArray *matlab::FmatrixToMx(const Fmatrix &f)
{
  // Allocate matlab matrix
  mwSize dims[2] = {f.Nrows, f.Ncols};
  mxArray *m = mxCreateNumericArray(2, dims, mxDOUBLE_CLASS, mxREAL);
  double *p = mxGetPr(m);

  // Fill
  for(int i = 0; i < f.Nrows; i++) {
    for(int j = 0; j < f.Ncols; j++) {
      p[i + f.Nrows*j] = f(i, j);
    }
  }

  return m;
}

EngineException::~EngineException() throw()
{
}

const char *EngineException::what() const throw()
{
  return "Matlab error";
}


matlab::Engine::Engine(bool echo)
{
  eng = engOpen("matlab -nosplash");
  if(eng == NULL) {
    throw EngineException();
  }

  this->echo = echo;

  // Construct output buffer
  outbuf_size = 8192;
  outbuf = new char[8192];

  for(int i = 0; i < outbuf_size; i++) {
    outbuf[i] = '\0';
  }
}

matlab::Engine::~Engine()
{
  int status = engClose(eng);
  if(status != 0) {
    throw EngineException();
  }

  delete [] outbuf;
}

void matlab::Engine::setEcho(bool _echo)
{
  echo = _echo;
}

void matlab::Engine::setVariable(const char *name, const FGmatrix &f)
{
  // Convert
  mxArray *m = FGmatrixToMx(f);

  // Pass to matlab engine
  int status = engPutVariable(eng, name, m);
  if(status != 0) {
    throw EngineException();
  }

  mxDestroyArray(m);
}

FGmatrix matlab::Engine::getVariable(const char *name)
{
  // Retrieve from matlab
  mxArray *mx = engGetVariable(eng, name);
  if(mx == NULL) {
    throw EngineException();
  }

  // Convert
  FGmatrix f = FGmatrixFromMx(mx);
  
  mxDestroyArray(mx);

  return f;
}

void matlab::Engine::execute(const std::string &code)
{
  if(echo) {
    engOutputBuffer(eng, outbuf, outbuf_size - 1);
    cout<<">> "<<code<<endl;
  }

  engEvalString(eng, code.c_str());

  if(echo) {
    // Skip a line + print output
    char *s;
    for(s = outbuf; s < outbuf + outbuf_size && *s != '\n'; s++);
    cout<<s; cout.flush();

    // Zero out output
    for(int i = 0; i < outbuf_size && outbuf[i] != '\0'; i++) {
      outbuf[i] = '\0';
    }
  }
}

void matlab::Engine::execute(const char *format, ...)
{
  va_list ap;
  char dummy_buf[1];

  // Allocate
  va_start(ap, format);
  int n = vsnprintf(dummy_buf, 0, format, ap);
  va_end(ap);

  char *s = new char[n+2];

  // Format
  va_start(ap, format);
  vsnprintf(s, n+1, format, ap);
  va_end(ap);

  // Run code
  execute(std::string(s));

  // Cleanup
  delete [] s;
}
