#ifndef __MATLAB_HPP__
#define __MATLAB_HPP__

#include <engine.h>
#include <mex.h>

#include "Fmatrix.hpp"
#include "FGmatrix.hpp"

#include <exception>
#include <string>
#include <cstdlib>

namespace matlab
{
  /**
   * Create an FGmatrix object from a Matlab C interface matrix.
   */
  FGmatrix FGmatrixFromMx(const mxArray *mx);

  /**
   * Create an FGmatrix object from a Matlab C interface matrix.
   */
  Fmatrix FmatrixFromMx(const mxArray *mx);

  /**
   * Convert an FGmatrix object to Matlab C interface matrix.
   *
   * @note Caller must free result with mxDestroyArray
   */
  mxArray *FGmatrixToMx(const FGmatrix &FG);

  /**
   * Convert an Fmatrix object to Matlab C interface matrix.
   *
   * @note Caller must free result with mxDestroyArray
   */
  mxArray *FmatrixToMx(const Fmatrix &F);


  class EngineException: public std::exception
  {
  public:
    virtual ~EngineException() throw ();
    virtual const char* what() const throw ();
  };

  /**
   * Engine with which to run Matlab code, interfaced with C.
   */
  class Engine
  {
  public:
    /**
     * Constructor, starts a matlab session
     */
    Engine(bool echo = false);
    
    /**
     * Destructor, ends the session and clears all data.
     */
    ~Engine();

    /**
     * Set a variable within the matlab session to the given matrix.
     */
    void setVariable(const char *name, const FGmatrix &value);

    /**
     * Get the value of a variable in the matlab in the form of an FGmatrix
     */
    FGmatrix getVariable(const char *name);

    /**
     * Execute an arbitrary line of Matlab code.
     */
    void execute(const std::string &code);

    /**
     * Execute Matlabe code constructed from a format string.
     */
    void execute(const char *format, ...);

    /**
     * Set whether the engine should print the matlab terminal to stdout.
     */
    void setEcho(bool);
    

  private:
    ::Engine *eng;
    bool echo;

    size_t outbuf_size;
    char *outbuf;

  };

}

#endif /* #ifndef __MATLAB_HPP__ */
