#include "FGmatrix.hpp"
#include "Lapack.hpp"

extern "C" {

  int dgeqrf_(integer *m, integer *n, doublereal *A, integer *lda,
	      doublereal *tau, doublereal *work, integer *lwork,
	      integer *info);

  int dorgqr_(integer *m, integer *n, integer *k, doublereal *A, integer *lda,
	      doublereal *tau, doublereal *work, integer *lwork, integer *info);

  int dgels_(char *trans, integer *m, integer *n, integer *nrhs,
	     doublereal *A, integer *lda,
	     doublereal *B, integer *ldb,
	     doublereal *work, integer *lwork,
	     integer *info);

  int dgelsy_(integer *m, integer *n, integer *nrhs,
	      doublereal *A, integer *lda,
	      doublereal *B, integer *ldb,
	      integer *jpvt, doublereal *rcond, integer *rank,
	      doublereal *work, integer *lwork,
	      integer *info);

  int dtrtrs_(char *uplo, char *trans, char *diag,
	      integer *n, integer *nrhs, doublereal *A, integer *lda,
	      doublereal *B, integer *ldb,
	      integer *info);

}

#define MIN(x, y) ((x) < (y) ? (x) : (y))
#define MAX(x, y) ((x) > (y) ? (x) : (y))

std::pair<FGmatrix, FGmatrix> FGmatrix::QR() const
{
  integer m = (integer)Nrows;
  integer n = (integer)Ncols;
  integer lda = m;
  doublereal *A = FGmatrixToLAPACK(*this);

  integer k = MIN(m, n);
  doublereal *tau = new double[k];

  // Get optimal work size
  doublereal worksize;
  integer lwork = -1, info;

  dgeqrf_(&m, &n, A, &lda, tau, &worksize, &lwork, &info);
  assert(info == 0);

  lwork = (integer)worksize;

  doublereal *work = new doublereal[lwork];

  // Calculate QR decomposition
  dgeqrf_(&m, &n, A, &lda, tau, work, &lwork, &info);
  assert(info == 0);

  /////////////////////// Output ///////////////////////

  // Build output R
  FGmatrix R(false, k, n);
  for(int i = 0; i < k; i++) {
    for(int j = 0; j < n; j++) {
      if(i <= j) { // Upper diagonal
	R(i, j) = A[lda*j + i];
      }
      else {
	R(i, j) = 0;
      }
    }
  }

  // Build output Q

  // Get optimal work size
  lwork = -1;
  dorgqr_(&m, &k, &k, A, &lda, tau, &worksize, &lwork, &info);
  assert(info == 0);
  lwork = (integer)worksize;

  // Build Q
  delete [] work;
  work = new doublereal[lwork];

  dorgqr_(&m, &k, &k, A, &lda, tau, work, &lwork, &info);
  assert(info == 0);

  FGmatrix Q = FGmatrixFromLAPACK(m, k, lda, A);

  // Cleanup
  delete [] work;
  delete [] tau;
  delete [] A;

  return std::pair<FGmatrix, FGmatrix>(Q, R);
}

FGmatrix FGmatrix::solveQR(const FGmatrix &b, double rcond_in) const
{
  assert(Nrows == b.Nrows);

  // Fill size variables
  integer m = (integer)Nrows;
  integer n = (integer)Ncols;
  integer nrhs = (integer)b.Ncols;

  integer lda = m;
  integer ldb = max(m, n);

  // Build matrices
  doublereal *A = FGmatrixToLAPACK(*this);
  doublereal *B = new doublereal[ldb * nrhs];
  FGmatrixFillLAPACK(b, ldb, B);

  // Pivoting
  doublereal rcond = (doublereal)rcond_in;
  integer *jpvt = new integer[n];
  integer rank;

  // Get optimal work size
  doublereal worksize;
  integer lwork = -1, info;

  dgelsy_(&m, &n, &nrhs, A, &lda, B, &ldb, jpvt, &rcond, &rank, &worksize, &lwork, &info);
  assert(info == 0);

  lwork = (integer)worksize;
  doublereal *work = new doublereal[lwork];

  // Call
  dgelsy_(&m, &n, &nrhs, A, &lda, B, &ldb, jpvt, &rcond, &rank, work, &lwork, &info);
  assert(info == 0);

  // Return value
  FGmatrix X = FGmatrixFromLAPACK(n, nrhs, ldb, B);

  // Cleanup
  delete [] work;
  delete [] jpvt;
  delete [] B;
  delete [] A;

  return X;
}

FGmatrix FGmatrix::solveQR(const FGmatrix &Q, const FGmatrix &R, const FGmatrix &b, double tol)
{
  assert(Q.Ncols == R.Nrows);
  assert(Q.Nrows == b.Nrows);

  // Cancel out Q
  FGmatrix QTb = (~Q) * b;

  // Solve Rx = QTb
  char uplo = 'U';
  char trans = 'N';
  char diag = 'N';

  integer lda = R.Nrows;
  integer n = R.Ncols;
  doublereal *A = FGmatrixToLAPACK(R);

  integer ldb = b.Nrows;
  integer nrhs = b.Ncols;
  doublereal *B = FGmatrixToLAPACK(b);

  assert(ldb >= n);

  integer info;
  dtrtrs_(&uplo, &trans, &diag, &n, &nrhs, A, &lda, B, &ldb, &info);
  assert(info == 0);

  // Output
  FGmatrix X = FGmatrixFromLAPACK(n, nrhs, ldb, B);

  // Cleanup
  delete [] B;
  delete [] A;

  return X;
}
