#include "sample.h"

#include <assert.h>

#ifndef MAX
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#endif

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "linalg.h"

struct sample
{
  gsl_rng *rng;
};

sample_t *sample_init(unsigned long int seed)
{
  sample_t *s;

  s = (sample_t *)malloc(sizeof(sample_t));

  // Initialize GSL random number generator
  s->rng = gsl_rng_alloc(gsl_rng_default);
  gsl_rng_set(s->rng, seed);

  return s;
}

void sample_end(sample_t *s)
{
  gsl_rng_free(s->rng);
  free(s);
}

double sample_uniform(sample_t *s, double min, double max)
{
  return gsl_ran_flat(s->rng, min, max);
}

size_t sample_discrete(sample_t *s, size_t k, const double *pi)
{
  double sum, x;
  size_t i;

  assert(k > 0);

  sum = 0;
  for(i = 0; i < k; i++) {
    sum += pi[i];
  }

  x = gsl_ran_flat(s->rng, 0, sum);

  for(i = 0; i < k && x >= 0; i++) {
    x -= pi[i];
  }

  return i-1;
}

static void sample_gaussian_sqrt(sample_t *s, size_t ndims, double *p, const double *mu, const double *sqrt_sigma, double *z)
{
  size_t i, j;

  for(i = 0; i < ndims; i++) {
    z[i] = gsl_ran_gaussian(s->rng, 1.0);
  }

  for(i = 0; i < ndims; i++) {
    p[i] = mu[i];
    for(j = 0; j < ndims; j++) {
      p[i] += sqrt_sigma[ndims*i + j] * z[j];
    }
  }
}

void sample_gaussian(sample_t *s, size_t ndims, size_t n, double *p, const double *mu, const double *sigma)
{
  double *sqrt_sigma, *z;
  size_t i;

  // Setup
  z = (double *)malloc(ndims * (ndims + 1) * sizeof(double));
  sqrt_sigma = z + ndims;
  linalg_sqrt(ndims, sqrt_sigma, sigma);

  for(i = 0; i < n; i++) {
    sample_gaussian_sqrt(s, ndims, p + ndims*i, mu, sqrt_sigma, z);
  }

  // Cleanup
  free(z);
}

void sample_MoG(sample_t *s,
                size_t ndims, size_t n, double *p,
                size_t k, const double *pi, const double *mu,
                const double *sigma)
{
  size_t i, l;
  unsigned int j;

  gsl_ran_discrete_t *label_dist;

  double *sqrt_sigma, *z;

  // Preprocess distributions
  label_dist = gsl_ran_discrete_preproc(k, pi);

  z = (double *)malloc((ndims + k * ndims * ndims) * sizeof(double));
  sqrt_sigma = z + ndims;
  for(i = 0; i < k; i++) {
    linalg_sqrt(ndims, sqrt_sigma + i*ndims*ndims, sigma + i*ndims*ndims);
  }

  // Sample points
  for(i = 0; i < n; i++) {
    l = gsl_ran_discrete(s->rng, label_dist);

    sample_gaussian_sqrt(s, ndims, p + ndims*i, mu + ndims*l, sqrt_sigma + ndims*ndims*l, z);
  }

  // Cleanup
  gsl_ran_discrete_free(label_dist);
}
